/*
 Copyright (c) 2009 Devin DeLong
 
 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
 files (the "Software"), to deal in the Software without
 restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following
 conditions:
 
 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.
 */

#pragma once
#ifndef BZU_FLAGINFO_H
#define BZU_FLAGINFO_H

// BZFS
#ifndef _BZFS_API_H_
#include "bzfsAPI.h"
#endif

// STL
#include <string>

#define TOTAL_FLAGS 48

/** Flag code constants.
 */
#define AGILITY "A"
#define BURROW "BU"
#define CLOAKING "CL"
#define CLOAKED_SHOT "CS"
#define GENOCIDE "G"
#define GUIDED_MISSILE "GM"
#define HIGH_SPEED "V"
#define IDENTIFY "ID"
#define INVISIBLE_BULLET "IB"
#define JUMPING "J"
#define LASER "L"
#define LOW_GRAVITY "LG"
#define MACHINE_GUN "MG"
#define MASQUERADE "MQ"
#define NARROW "N"
#define OSCILLATION_OVERTHRUSTER "OO"
#define PHANTOM_ZONE "PZ"
#define QUICK_TURN "QT"
#define RAPID_FIRE "F"
#define RICOCHET "R"
#define SEER "SE"
#define SHIELD "SH"
#define SHOCKWAVE "SW"
#define SUPPER_BULLET "SB"
#define STEALTH "ST"
#define STEAM_ROLLER "SR"
#define THIEF "TH"
#define TINY "T"
#define USELESS "US"
#define WINGS "WG"
#define BLINDNESS "B"
#define BOUNCY "BY"
#define COLORBLINDNESS "CB"
#define FORWARD_ONLY "FO"
#define JAMMING "JM"
#define LEFT_TURN_ONLY "LT"
#define MOMENTUM "M"
#define NO_JUMPING "NJ"
#define OBESITY "O"
#define RIGHT_TURN_ONLY "RT"
#define REVERSE_ONLY "RO"
#define REVERSE_CONTROLS "RC"
#define TRIGGERHAPPY "TR"
#define WIDE_ANGLE "WA"
#define RED_TEAM "R*"
#define GREEN_TEAM "G*"
#define BLUE_TEAM "B*"
#define PURPLE_TEAM "P*"
#define PROXIMITY_MINE "PM"


namespace bzu
{
// FlagInfo class
//------------------------------------------------------------------------------//

/*
 NOTE: flag codes are case-sensitive 
 */

/*!
    @class
    @abstract    This class is used to get information about flags in BZFlag.
*/

class FlagInfo {
    
  // FlagInfo::FlagInfoData class
  //------------------------------------------------------------------------------//
  
  /*!
   @class
   @abstract    A small class that stores properties of a flag.
   @discussion  This class is only used internally by the FlagInfo 
   class and, as such, should probably be removed from
   this header.
   */
  class FlagInfoData {
    
  public:
    // There will not be a default constructor.
    /*!
     @method     
     @abstract   This is the only constructor for this class.
     @discussion
     @param name The flag code.
     @param shotType The default shot type associated with the flag.
     @param quality Indicates whether this flag is good or bad.
     @param team Indicates whether or not the flag is a team flag, and
     which team flag it is.
     @param fullName The full name of the flag.
     */
    FlagInfoData(const std::string& name,
                 bz_eShotType shotType,
                 bz_eFlagQuality quality,
                 bz_eTeamType team,
                 const std::string& fullName)
    : _name(name), 
    _shotType(shotType), 
    _quality(quality), 
    _team(team),
    _fullName(fullName)
    {
      
    }
    
    /*!
     @method     
     @abstract Returns the flag code.
     */
    const std::string& getName() const;
    
    /*!
     @method     
     @abstract Returns the flag's full name.
     */
    const std::string& getFullName() const;
    
    /*!
     @method     
     @abstract Returns the shot type.
     */
    bz_eShotType getShotType() const; 
    
    /*!
     @method     
     @abstract Returns the quality of the flag: a good flag or bad flag.
     */
    bz_eFlagQuality getQuality() const; 
    
    /*!
     @method     
     @abstract Returns the team associated with the flag.
     @discussion If the flag is not a team flag, this returns eNoTeam.
     */
    bz_eTeamType getTeam() const;
    
  private:
    
    // The flag code.
    std::string _name;  
    // The shot type.
    bz_eShotType _shotType;
    // The flag quality.
    bz_eFlagQuality _quality;
    // The flag team.
    bz_eTeamType _team;
    // The full flag name
    std::string _fullName;
  };
    
public:
  
  /*!
   @method     
   @abstract Default-initializes everything.
   */
  FlagInfo();
  
  /*!
   @method     
   @abstract Initializes everything.
   @discussion This constructor determines whether or not the flag
     passed is a valid flag or not and initializes FlagInfo's 
     internal state.  After constructing this class, users 
     should call isValid() to see if the specified flag is a 
     valid flag in BZFlag.
   @param flag The flag code of the flag.
   */
  explicit FlagInfo(const std::string& flag);
  
  // Construct from a flag ID
  /*!
   @method     
   @abstract This constructor constructs a FlagInfo class from an integer
             flag ID used in BZFlag.
   @discussion This constructor determines whether or not the flag
     passed is a valid flag or not and initializes FlagInfo's 
     internal state.  After constructing this class, users 
     should call isValid() to see if the specified flag is a 
     valid flag in BZFlag.
   @param ID An integer flag ID used in BZFlag.
   */
  explicit FlagInfo(int ID);
  
  /*!
   @method     
   @abstract Indicates if the flag is valid or not.
   */
  bool isValid() const { return _valid; }
  
  /*!
   @method     
   @abstract This function reinitializes a FlagInfo class with a new flag.
   @param flag The flag code of a flag.
   */
  void set(const std::string& flag);
  
  /*!
   @method     
   @abstract This function reinitializes a FlagInfo class with a new flag.
   @param flag The ID of a flag.
   */
  void set(int ID);
  
  /*!
   @method     
   @abstract This function returns the default shot type 
             associated with the current flag.
   */
  bz_eShotType getShotType() const 
    { return _validFlags[_index].getShotType(); }
  
  /*!
   @method     
   @abstract This function returns the quality of the current flag.
   */
  bz_eFlagQuality getQuality() const 
    { return _validFlags[_index].getQuality(); }  
  
  /*!
   @method     
   @abstract This function returns the team associated with the current flag.
   */
  bz_eTeamType getTeam() const 
    { return _validFlags[_index].getTeam(); }
  
  /*!
   @method     
   @abstract This function returns the flag code of the current flag.
   */
  const std::string& getName() const 
    { return _validFlags[_index].getName(); }
  
  /*!
   @method     
   @abstract This function returns the full name of the current flag.
   */
  const std::string& getFullName() const 
    { return _validFlags[_index].getFullName(); }
  
  // Convenience functions
  /*!
   @method     
   @abstract Returns true if the current flag is a good flag.
   */
  bool isGood() const;
  
  /*!
   @method     
   @abstract Returns true if the current flag is a bad flag.
   */
  bool isBad() const;
  
  /*!
   @method     
   @abstract Returns true if the current flag is a team flag.
   */
  bool isTeam() const;
  
  /*!
   @method     
   @abstract Returns true if the current flag is a super flag.
   */  
  bool isSuper() const;
  
  /*!
   @method     
   @abstract Returns true if the current flag has the standard shot type.
   */
  bool hasStandardShot() const;
  
  /*!
   @method     
   @abstract Returns true if the current flag has a nonstandard shot type.
   */
  bool hasNonStandardShot() const;
  
  // Optimized static functions
  /*!
   @method     
   @abstract Returns true if the current flag is a good flag.
   @discussion This function uses an optimized search algorithm
     that is faster than constructing a FlagInfo class, as
     the number of flags to check against has been narrowed down
     to the good flags.  The constructors check against every 
     known flag.  Although, if you need to get more information
     than what this function provides, it may be best to construct
     a FlagInfo class; it may reduce searching, as these static 
     functions have to refind the flag everytime they are called.
   @param flag The flag code of the flag.
   */
  static bool isGood(const std::string& flag);
  
  /*!
   @method     
   @abstract Returns true if the current flag is a bad flag.
   @discussion This function uses an optimized search algorithm
     that is faster than constructing a FlagInfo class, as
     the number of flags to check against has been narrowed down
     to the bad flags.  Although, if you need to get more information
     than what this function provides, it may be best to construct
     a FlagInfo class; it may reduce searching, as these static 
     functions have to refind the flag everytime they are called.
   @param flag The flag code of the flag.
   */
  static bool isBad(const std::string& flag);
  
  /*!
   @method     
   @abstract Returns true if the current flag is a team flag.
   @discussion This function uses an optimized search algorithm
     that is faster than constructing a FlagInfo class, as
     the number of flags to check against has been narrowed down
     to the team flags.  Although, if you need to get more information
     than what this function provides, it may be best to construct
     a FlagInfo class; it may reduce searching, as these static 
     functions have to refind the flag everytime they are called.
   @param flag The flag code of the flag.
   */
  static bool isTeam(const std::string& flag);
  
  /*!
   @method     
   @abstract Returns true if the current flag is a super flag.
   @discussion This function uses an optimized search algorithm
     that is faster than constructing a FlagInfo class, as
     the number of flags to check against has been narrowed down
     to the super flags.  Although, if you need to get more information
     than what this function provides, it may be best to construct
     a FlagInfo class; it may reduce searching, as these static 
     functions have to refind the flag everytime they are called.
   @param flag The flag code of the flag.
   */ 
  static bool isSuper(const std::string& flag);
  
  /*!
   @method     
   @abstract Returns true if the current flag has the standard shot type.
   @discussion This function uses an optimized search algorithm
     that is faster than constructing a FlagInfo class, as
     the number of flags to check against has been narrowed down
     to the flags with standard shots.  Although, if you need to get more information
     than what this function provides, it may be best to construct
     a FlagInfo class; it may reduce searching, as these static 
     functions have to refind the flag everytime they are called.
   @param flag The flag code of the flag.
   */
  static bool hasStandardShot(const std::string& flag);
  
  /*!
   @method     
   @abstract Returns true if the current flag has a nonstandard shot type.
   @discussion This function uses an optimized search algorithm
     that is faster than constructing a FlagInfo class, as
     the number of flags to check against has been narrowed down
     to the flags with nonstandard shots.  Although, if you need to get more information
     than what this function provides, it may be best to construct
     a FlagInfo class; it may reduce searching, as these static 
     functions have to refind the flag everytime they are called.
   @param flag The flag code of the flag.
   */
  static bool hasNonStandardShot(const std::string& flag);
  
private:
  
  // Searching
  std::size_t find(const std::string& flag) const;
  std::size_t find(int ID) const;
  
  // Yep, static
  static std::size_t find(const std::string& flag, 
                          std::size_t begin, 
                          std::size_t end);
  
private:
  // The validity of the flag
  bool _valid;
  // Stores the index of the flag
  std::size_t _index;
  // An array of valid flags
  const static FlagInfoData _validFlags[TOTAL_FLAGS];
};



//------------------------------------------------------------------------------//

inline FlagInfo::FlagInfo() : _valid(false), _index(0)
{
  
}


// FlagInfo::FlagInfoData functions
//------------------------------------------------------------------------------//

inline const std::string& FlagInfo::FlagInfoData::getName() const
{
  return _name;
}

inline bz_eShotType FlagInfo::FlagInfoData::getShotType() const
{
  return _shotType;
}

inline bz_eFlagQuality FlagInfo::FlagInfoData::getQuality() const
{
  return _quality;
}

inline bz_eTeamType FlagInfo::FlagInfoData::getTeam() const
{
  return _team;
}

inline const std::string& FlagInfo::FlagInfoData::getFullName() const
{
  return _fullName;
}

}



// The End
//------------------------------------------------------------------------------//


#endif /* BZU_FLAGINFO_H */

